/* file: "os.c" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

/*
This module implements the operating system specific routines including:

  - user and timer interrupt handling
  - dynamic loading
  - shell command
  - elapsed user time
  - filename expansion
*/

#define ___VERSION 9
#include "gambit.h"
___BEGIN_MODULE

/*---------------------------------------------------------------------------*/

/* Select OS routines to use. */

#ifdef sparc
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef sgi
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef __alpha
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef hpux
#define USE_UNIX
#define USE_shl_load
#define USE_sigaction
#endif

#ifdef _AIX
#define USE_UNIX
#define USE_sigaction
#endif

#ifdef linux
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif
#endif

#ifdef _MSC_VER
#define USE_WINDOWS
#endif

#ifdef __WATCOMC__
#ifdef __386__
#define USE_WINDOWS
#endif
#endif

#ifdef __MWERKS__
#endif

#ifdef __MACHTEN_PPC__
#endif

#ifdef MACHTEN
#endif

/*---------------------------------------------------------------------------*/

/* Determine which header files to include. */

#ifdef USE_UNIX
#define USE_times
#define INCLUDE_unistd_h
#define INCLUDE_pwd_h
#define INCLUDE_sys_types_h
#define INCLUDE_sys_time_h
#define INCLUDE_sys_times_h
#define INCLUDE_time_h
#endif

#ifdef USE_dlopen
#define INCLUDE_dlfcn_h
#endif

#ifdef USE_shl_load
#define INCLUDE_dl_h
#endif

#ifdef USE_sigaction
#define INCLUDE_signal_h
#endif

#ifdef USE_signal
#define INCLUDE_signal_h
#endif

#ifdef USE_WINDOWS
#define USE_LoadLibrary
#define INCLUDE_windows_h
#define INCLUDE_direct_h
#endif

/*---------------------------------------------------------------------------*/

/* Inclusion of header files. */

#include <stdlib.h>
#include <string.h>

#include "os.h"
#include "setup.h"

#ifdef INCLUDE_unistd_h
#include <unistd.h>
#endif

#ifdef INCLUDE_pwd_h
#include <pwd.h>
#endif

#ifdef INCLUDE_dlfcn_h
#include <dlfcn.h>
#endif

#ifdef INCLUDE_dl_h
#include <dl.h>
#endif

#ifdef INCLUDE_sys_types_h
#include <sys/types.h>
#endif

#ifdef INCLUDE_sys_time_h
#include <sys/time.h>
#endif

#ifdef INCLUDE_sys_times_h
#include <sys/times.h>
#endif

#ifdef INCLUDE_time_h
#include <time.h>
#endif

#ifdef INCLUDE_signal_h
#include <signal.h>
#endif

#ifdef INCLUDE_windows_h
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#ifdef INCLUDE_direct_h
#include <direct.h>
#endif

/*---------------------------------------------------------------------------*/

/* User and timer interrupt handling. */

#ifndef USE_sigaction

void ___set_timer_interval (msecs)
int msecs;
{
}

static void init_interrupt_handlers ()
{
}

static void disable_interrupts ()
{
}

static void enable_interrupts()
{
}

#endif

#ifdef USE_sigaction

void ___set_timer_interval (msecs)
int msecs;
{
  struct itimerval virtual_tv;
  int secs = msecs/1000;
  int usecs = (msecs%1000)*1000;
  virtual_tv.it_interval.tv_sec  = secs;
  virtual_tv.it_interval.tv_usec = usecs;
  virtual_tv.it_value.tv_sec     = secs;
  virtual_tv.it_value.tv_usec    = usecs;
  setitimer (ITIMER_VIRTUAL, &virtual_tv, 0);
}

static void user_interrupt_handler ()
{
  ___raise_interrupt (0);
}

static void timer_interrupt_handler ()
{
  ___raise_interrupt (1);
}

static void init_interrupt_handlers ()
{
  struct sigaction act;

  act.sa_handler = user_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (SIGINT, &act, 0);

  act.sa_handler = timer_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (SIGVTALRM, &act, 0);

  ___set_timer_interval (___DEFAULT_TIMER_INTERVAL);
}

static void disable_interrupts ()
{
  sigset_t mask;
  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigaddset (&mask, SIGVTALRM);
  sigprocmask (SIG_BLOCK, &mask, 0);
}

static void enable_interrupts()
{
  sigset_t mask;
  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigaddset (&mask, SIGVTALRM);
  sigprocmask (SIG_UNBLOCK, &mask, 0);
}

#endif

/*---------------------------------------------------------------------------*/

/* Dynamic loading. */

void *___dynamic_load (path, errmsg)
char *path;
char **errmsg;
{
  void *result = 0;
  char *module_linker = "___init_module";

#ifndef USE_dlopen
#ifndef USE_shl_load
#ifndef USE_LoadLibrary

  *errmsg = "Dynamic loading is not available on this platform";
  result = 0;

#endif
#endif
#endif

#ifdef USE_dlopen

#ifdef RTLD_NOW
  void *descr = dlopen (path, RTLD_NOW);
#else
  void *descr = dlopen (path, 1);
#endif
  if (descr != 0)
    result = dlsym (descr, module_linker);
  if (result == 0)
    *errmsg = (char *)dlerror();

#endif

#ifdef USE_shl_load

  shl_t descr = shl_load (path, BIND_IMMEDIATE, 0);
  if (descr != 0)
    if (shl_findsym (&descr, module_linker, TYPE_PROCEDURE, &result))
      result = 0;
  if (result == 0)
    *errmsg = "unknown error";

#endif

#ifdef USE_LoadLibrary

  HANDLE descr = LoadLibrary (path);
  if (descr != 0)
    result = GetProcAddress (descr, module_linker);
  if (result == 0)
    *errmsg = "unknown error";

#endif

  return result;
}

/*---------------------------------------------------------------------------*/

/* Shell command. */

int ___shell_command (cmd)
char *cmd;
{
  int result;

  disable_interrupts ();
  result = system (cmd);
  enable_interrupts ();

  return result;
}

/*---------------------------------------------------------------------------*/

/* Elapsed user time. */

#ifdef USE_TickCount

static unsigned long initial_tickcount;

static void init_user_time ()
{
  initial_tickcount = (unsigned long)TickCount();
}

#else

static void init_user_time ()
{
}

#endif

unsigned long ___user_time ()
{

#ifndef USE_time
#ifndef USE_times
#ifndef USE_clock
#ifndef USE_getrusage
#ifndef USE_TickCount

  return 0;

#endif
#endif
#endif
#endif
#endif

#ifdef USE_time

  time_t t;
  time (&t);
  return t * 1000;
}

#endif

#ifdef USE_times

#ifdef CLK_TCK
#define CLOCK_TICKS_PER_SEC CLK_TCK
#else
#ifdef _SC_CLOCKS_PER_SEC
#define CLOCK_TICKS_PER_SEC sysconf(_SC_CLOCKS_PER_SEC)
#else
#ifdef _SC_CLK_TCK
#define CLOCK_TICKS_PER_SEC sysconf(_SC_CLK_TCK)
#else
@error Cannot find a definition for CLOCK_TICKS_PER_SEC
#endif
#endif
#endif

  struct tms t;
  times (&t);
  return t.tms_utime * 1000 / CLOCK_TICKS_PER_SEC;

#endif

#ifdef USE_clock

  return clock () * 1000 / CLOCKS_PER_SEC;

#endif

#ifdef USE_getrusage

  struct rusage ru;
  getrusage (RUSAGE_SELF, &ru);
  return ru.ru_utime.tv_sec * 1000 + ru.ru_utime.tv_usec / 1000;

#endif

#ifdef USE_TickCount

  return ((unsigned long)TickCount () - initial_tickcount) * 50 / 3;

#endif
}

/*---------------------------------------------------------------------------*/

/* Filename expansion. */

#ifdef USE_WINDOWS

#define DIR_SEPARATOR1 '\\'
#define DIR_SEPARATOR2 '/'
#define DRIVE_SEPARATOR ':'

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1 || (c) == DIR_SEPARATOR2)
#define SEPARATOR(c)(DIR_SEPARATOR(c) || (c) == DRIVE_SEPARATOR)

static int normalize_dir_path (dir_path, norm_dir_path, max_size)
char *dir_path, *norm_dir_path;
long max_size;
{
  char cwd1[___PATH_MAX_SIZE+1];
  int result = 0;
  if (getcwd (cwd1, ___PATH_MAX_SIZE+1) != 0)
    {
      long i = 0;
      while (dir_path[i] != '\0' && !SEPARATOR(dir_path[i]))
        i++;
      if (dir_path[i] == DRIVE_SEPARATOR && dir_path[i+1] != '\0')
        {
          char cwd2[___PATH_MAX_SIZE+1];
          char tmp = dir_path[i+1];
          dir_path[i+1] = '\0';
          if (chdir (dir_path) == 0)
            {
              if (getcwd (cwd2, ___PATH_MAX_SIZE+1) != 0)
                {
                  dir_path[i+1] = tmp;
                  if (chdir (dir_path) == 0)
                    {
                      result = getcwd (norm_dir_path, max_size) != 0;
                      chdir (cwd2);
                    }
                }
              chdir (cwd1);
            }
        }
      else if (chdir (dir_path) == 0)
        {
          result = getcwd (norm_dir_path, max_size) != 0;
          chdir (cwd1);
        }
    }
  return result;
}

#else

#define DIR_SEPARATOR1 '/'

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1)
#define SEPARATOR(c)DIR_SEPARATOR(c)

static int normalize_dir_path (dir_path, norm_dir_path, max_size)
char *dir_path, *norm_dir_path;
long max_size;
{
  char cwd[___PATH_MAX_SIZE+1];
  int result = 0;
  if (getcwd (cwd, ___PATH_MAX_SIZE+1) != 0 && chdir (dir_path) == 0)
    {
      result = getcwd (norm_dir_path, max_size) != 0;
      chdir (cwd);
    }
  return result;
}

#endif

static char root_path[] = { DIR_SEPARATOR1, '\0' };

int ___absolute_path (path, abs_path, max_size)
char *path, *abs_path;
long max_size;
{
  int result = 0;
  char tmp[___PATH_MAX_SIZE+1];
  long i = 0;
  long j = 0;
  long k;

  disable_interrupts ();

  if (path[0] == '\0')
    if (getcwd (abs_path, max_size) == 0)
      goto ret;
    else
      goto terminal_dir_separator;

  if (path[0] == '~')
    {
      char *home;

      i++;
      while (path[i]!='\0' && !DIR_SEPARATOR(path[i]) &&
             i <= ___PATH_MAX_SIZE+1)
        {
          tmp[i-1] = path[i];
          i++;
        }

      if (i > ___PATH_MAX_SIZE+1)
        goto ret;

      if (i == 1)
        {
          home = getenv ("HOME");
          if (home == 0)
            home = root_path;
        }
      else
#ifdef USE_UNIX
        {
          struct passwd *p;
          tmp[i-1] = '\0';
          p = getpwnam (tmp);
          if (p == 0)
            goto ret;
          home = p->pw_dir;
          if (home == 0)
            goto ret;
        }
#else
        goto ret;
#endif

      j = 0;
      while (*home != '\0' && j < ___PATH_MAX_SIZE+1)
        tmp[j++] = *home++;
    }

  k = i;

  while (path[i] != '\0' && j < ___PATH_MAX_SIZE+1)
    tmp[j++] = path[i++];

  if (j >= ___PATH_MAX_SIZE+1)
    goto ret;

  tmp[j] = '\0';

  if (normalize_dir_path (tmp, abs_path, max_size))
    {
      terminal_dir_separator:
      j = strlen (abs_path);
      if (j > 0 && !DIR_SEPARATOR(abs_path[j-1]))
        {
          abs_path[j++] = DIR_SEPARATOR1;
          if (j >= max_size)
            goto ret;
          abs_path[j] = '\0';
        }
    }
  else
    {
      i = strlen (path);
      while (i > k && !SEPARATOR(path[i]))
        {
          i--;
          j--;
        }

      if (SEPARATOR(path[i]))
        {
          i++;
          j++;
          if (j >= ___PATH_MAX_SIZE+1)
            goto ret;
          tmp[j] = '\0';
          if (!normalize_dir_path (tmp, abs_path, max_size))
            goto ret;
        }
      else
        {
          if (getcwd (abs_path, max_size) == 0)
            goto ret;
        }

      j = strlen (abs_path);
      if (!DIR_SEPARATOR(abs_path[j-1]))
        abs_path[j++] = DIR_SEPARATOR1;

      while (path[i] != '\0' && j < max_size)
        abs_path[j++] = path[i++];

      if (j >= max_size)
        goto ret;

      abs_path[j] = '\0';
    }

  result = 1;

  ret:

  enable_interrupts ();

  return result;
}

/*---------------------------------------------------------------------------*/

void ___init_os ()
{
  init_interrupt_handlers ();
  init_user_time ();
}
